(* ::Package:: *)

BeginPackage["CompareWeights`"];

<<MaTeX`;
CompareWeights::usage = "Compares balance heuristic, power heuristic and optimal weighting";

l2Approx::usage = "Produces L2 approximation by means of sampling strategies. For usage with methods CompareWeights producing maps, such as CompareWeightsPaper.";
pltTable::usage = "Produces Grid of graphs for a paper. For usage with methods CompareWeights producing maps, such as CompareWeightsPaper.";
pltProd::usage = "Produces graphs sampling densities products. For usage with methods CompareWeights producing maps, such as CompareWeightsPaper.";
pltProdL2::usage = "Produces graphs sampling densities products. For usage with methods CompareWeights producing maps, such as CompareWeightsPaper.";
pltVar::usage = "";
pltAANA::usage = "todo";
mDot::usage = ".... For usage with methods CompareWeights producing maps, such as CompareWeightsPaper.";
mDot2::usage = ".... For usage with methods CompareWeights producing maps, such as CompareWeightsPaper.";
mDot3::usage = ".... For usage with methods CompareWeights producing maps, such as CompareWeightsPaper.";
varBH::usage = ".... For usage with methods CompareWeights producing maps, such as CompareWeightsPaper.";
rejectionSample::usage = ".... For usage with methods CompareWeights producing maps, such as CompareWeightsPaper.";
abSystemEstimate::usage = ".... For usage with methods CompareWeights producing maps, such as CompareWeightsPaper.";

Begin["`Private`"];


CompareWeights[f_,x_,xMin_,xMax_,n_,pUnorm_,pRndGen_]:= 
Module[{
	xrange, yrange, nzxrange, nzyrange,
	G, pB, pL, p, pLabels, nB, nL, ff, paux, pV, V, H, normaln, viewn, legends, Le, pLe, pVe, pU, result, veachBound,
	wBalB, wBalL, wPowB, wPowL, wOptB, wOptL,
	(* Function computing one term of variance corresponding to the given wi, pi, ni *)
	computeVar, computeMean,
	(* Lists of weight definitions, sums, short names, long names, labels, means and variances *)
	addWeight, wMeans = {}, wVars = {}, wDefs={}, wSums={}, wSNames={}, wLNames={}, wLabels={}, wIdx,
	(* linear system *)
	A, b, \[Alpha],
	(* Integrator settings *)	
	eps, acc, prec, wprec, method
},
(* Pdfs *)
p=pUnorm/Integrate[pUnorm,{x,xMin,xMax}];
pLabels=Map[("p"<>ToString[#])&,Range[Length[p]]];

(* The target integral *)
result=Integrate[f,{x,xMin,xMax}];

(* Bound on balance heuristic variance given by Veach *)
veachBound=N[(1/Min[n]-1/Total[n])*result^2];

(* Function computing one term of mean corresponding to the given wi, pi, ni *)
computeMean[wi_,pi_,ni_]:=NIntegrate[Simplify[wi*f],{x,xMin,xMax}];

(* Function computing one term of variance corresponding to the given wi, pi, ni *)
computeVar[wi_,pi_,ni_]:=1/ni*(NIntegrate[Simplify[wi^2*f^2/pi],{x,xMin,xMax}]-NIntegrate[Simplify[wi*f],{x,xMin,xMax}]^2);

(* Appends the given weight to the lists above *)
addWeight[nameLong_,nameShort_,definition_]:=(
	wDefs=Append[wDefs,definition];
	wSums=Append[wSums,Total[definition]];
	wSNames=Append[wSNames,nameShort];
	wLNames=Append[wLNames,nameLong];
	wLabels=Append[wLabels,Map[("w"<>ToString[#]<>nameShort)&,Range[Length[p]]]];
	wMeans=Append[wMeans,Total[MapThread[computeMean,{definition,p,n}]]];
	wVars=Append[wVars,Total[MapThread[computeVar,{definition,p,n}]]];
	wIdx[nameShort]=Length[wDefs];
);

(* Single techniques only (no combination) *)
Quiet[Map[(addWeight["p"<>ToString[#]<>" only", "only"<>ToString[#], UnitVector[Length[p],#]])&,Range[Length[p]]]];

(* Maximum heuristic *)
addWeight["maximum heuristic", "max", ((If[#<Max[n*p],0,#])&/@(n*p))/Total[(If[#<Max[n*p],0,#])&/@(n*p)]];

(* Cutoff heuristic (\[Alpha] = 0.1) *)
addWeight["cutoff heuristic", "cut", ((If[#<0.1*Max[n*p],0,#])&/@(n*p))/Total[(If[#<0.1*Max[n*p],0,#])&/@(n*p)]];

(* Balance heuristic *)
addWeight["balance heuristic", "bal", n*p/Total[p*n]];

(* Power heuristic with exponent 2 *)
addWeight["power^2 heuristic", "pow", (n*p)^2/Total[(p*n)^2]];

(* Optimal weighting - precise *)
A=Table[NIntegrate[p[[i]]*p[[j]]/Total[p*n],{x,xMin,xMax}],{i,Length[p]},{j,Length[p]}];
b=Table[NIntegrate[f*p[[i]]/Total[p*n],{x,xMin,xMax}],{i,Length[p]}];
\[Alpha]=LinearSolve[A,b];
addWeight["optimal weighting", "opt", wDefs[[wIdx["bal"]]]*(1-Total[\[Alpha]*p]/f)+\[Alpha]*p/f];

(* Output - variance *)
Print[
	Grid[
		Join[
		 {{"technique","mean","variance","bal-this (<= "<>ToString[veachBound]<>")","this/min"}},
		 Table[{wLNames[[i]],wMeans[[i]],wVars[[i]],wVars[[wIdx["bal"]]]-wVars[[i]],wVars[[i]]/Min[wVars]},{i,Length[wDefs]}]
		],
		Frame->All
		(*ItemStyle->{
			Automatic, Automatic, {{{1,1},{1,Length[tableHead[[1]]]}}->Bold}
		}*)
	]
];

<|
	"A" -> A, "b" -> b, "alpha" -> \[Alpha],
	"wMeans" -> wMeans, "wVars" -> wVars, "wDefs" -> wDefs, "wSums" -> wSums, "wSNames" -> wSNames, "wLNames" -> wLNames, "wLabels" -> wLabels, "wIdx" -> wIdx,
	"n"-> n, "f" -> f, "p" -> p, "pRndGen" -> pRndGen, "pLabels" -> pLabels, "result" -> result, "veachBound" -> veachBound,
	"range"-> {x, xMin, xMax}
|>

(*(* Output - optimal weights at endpoints *)
Print["Optimal weights at the endpoints:"];
wTableHead={{"weight", "val at "<>ToString[N[xMin]], ToString[N[xMin]]<>" \[LeftArrow] lim", "lim \[Rule] "<>ToString[N[xMax]], "val at "<>ToString[N[xMax]]}};
wTableData=Quiet[Table[{"w"<>ToString[i], wDefs[[wIdx["opt"]]][[i]]/.x->xMin, Limit[wDefs[[wIdx["opt"]]][[i]],x->xMin,Direction->-1], Limit[wDefs[[wIdx["opt"]]][[i]],x->xMax,Direction->1],wDefs[[wIdx["opt"]]][[i]]/.x->xMax},{i,Length[p]}]];
Print[Grid[Join[wTableHead,wTableData],Frame->All,ItemStyle->{Automatic,Automatic,{{{1,1},{1,Length[wTableHead[[1]]]}}->Bold}}]];

(* Output - plots *)
plotInp=Plot[Evaluate@Join[p,{f}],{x,xMin,xMax},PlotStyle->Thick,PlotLabels->Placed[Join[pLabels,{"f"}],Above]];
plotApprox=Plot[{Total[\[Alpha] p], f},{x,xMin,xMax},PlotStyle->Thick,PlotLabels->Placed[{"approx", "f"},Above]];
plotW[wIndex_]:=Plot[Evaluate@Join[wDefs[[wIndex]],If[debug,{wSums[[wIndex]]},{}]],{x,xMin,xMax},PlotStyle->Thick,PlotLabels->Placed[Join[wLabels[[wIndex]],If[debug,{"sum"},{}]],Above]];
Print[GraphicsGrid[{Join[{plotInp, plotApprox},Map[plotW,Range[Length[p]+1,Length[wDefs]]],If[estimateOpt,{plotOptEst},{}]]},ImageSize->Full]];*)
]


l2Approx[a31_] := Module[{A, b, \[Alpha] , p,  n, f, range},
   (* Optimal weighting - precise *)
   p := a31["p"];
   n := a31["n"];
   f := a31["f"];
   range := a31["range"];
   (*Print[{p, n, f, range}];*)
   A = Table[NIntegrate[p[[i]]*p[[j]], range], {i, Length[p]}, {j, Length[p]}];
   b = Table[NIntegrate[f*p[[i]], range], {i, Length[p]}];
   
   LinearSolve[A, b]
]


pltTable[a11_] := Module[
{
  idx, MTex, pStyles, colors
},
  MTex[x_]:= MaTeX[x, Preamble->{"\\usepackage[libertine]{newtxmath}","\\usepackage{xcolor}"}, FontSize->8, ContentPadding->True];
  pStyles = Join[
  {RGBColor[#1], Thick}&/@ StringSplit["#E69F00 #56B4E9 #009E73 #F0E442 #0072B2"][[1;;Length[a11["p"]]]],
  {{Black, Dashed}}
  ];
  idx = Join[
    Ordering[ a11["wVars"][[ 1;;Length[a11["p"]] ]], 1],
    Range[Length[a11["wDefs"]]-Length[a11["p"]]] + Length[a11["p"]]
  ];
  colors = {"E69F00", "56B4E9", "009E73", "F0E442", "0072B2"};
  Grid[{
      {
         GraphicsGrid[
           {
             Show[#1, AspectRatio -> 1/2]& /@
             Join @@ {{
               Plot[Evaluate@Join[a11["p"], {a11["f"]}], a11["range"],
                 PlotStyle -> pStyles,
                 PlotLabels -> Placed[Join[ MTex["\\textcolor[HTML]{" <> colors[[#1]] <> "}{p_" <> ToString[#1] <> "}"]& /@ Range[Length[a11["p"]]], MTex[{"f"}]], Above],
                 PlotLabel -> " ",
                 PlotRange -> All]
               },
               Table[
                 Plot[Evaluate@a11["wDefs"][[i]], a11["range"],
                   PlotStyle -> pStyles, 
                   PlotLabel -> Map["Var: " <> ToString[NumberForm[#1, 3]] &, a11["wVars"]][[i]],
                   PlotLabels-> Placed[ MTex["\\textcolor[HTML]{" <> colors[[#1]] <> "}{w_" <> ToString[#1] <> "}"]& /@ Range[Length[a11["p"]]], Above],
                   PlotRangeClipping->False
                 ],
                 {i, idx}
               ],
               {
               Plot[
                 Evaluate@{
                   a11["f"] - Total[a11["alpha"] a11["p"]], 
                   a11["f"] - Total[l2Approx[a11] a11["p"]], 
                   Total[a11["n"] a11["p"]]/Total[a11["n"]]
                  }, 
                 a11["range"],
                 PlotStyle -> {Thick, Thick, Dashed},
                 PlotLabels -> MapThread[Placed[#1, #2] &, {{"opt", "l2", "effective mix"}, {Above, {Above}, Above}}],
                 PlotLabel -> " ",
                 PlotRange -> Automatic
               ]
               }}
           },
           Spacings-> 0, 
           ImageSize -> Full,
           AspectRatio -> 1/5
         ], SpanFromLeft, SpanFromLeft, SpanFromLeft, SpanFromLeft
      },
      {"  a)", "  b)", "  c)", "  d)" , "  e)" , "  f)" , "  g)"}
    },
    (*Alignment\[Rule]{{Left,Left,Left}},*)
    (*ItemSize\[Rule]{{20, 5, 10}, {5, 5, 5}},*)
    ItemSize->12,
    Frame -> None
  ]
]


pltProd[a_] := Module[{},
    GraphicsGrid[
      {
       Show[#1, AspectRatio -> 1/2, PlotRange->All, PlotStyle->Thick]& /@
       {
        Plot[Evaluate@{a["p"] a["f"]/Total[a["p"]]}, a["range"], 
         PlotRange->Full,
         PlotStyle->Thick
        ],
        Plot[Evaluate@{Transpose[{a["p"]}].{a["p"]}/Total[a["p"]]}, a["range"], 
         PlotRange->Full,
         PlotStyle->Thick
        ]
      }},
      Spacings-> 0, 
      ImageSize -> Full,
      AspectRatio -> 1/2
    ]
]


pltProdL2[a_] := Module[{},
    GraphicsGrid[
      {
       Show[#1, AspectRatio -> 1/2, PlotRange->Automatic]& /@
       {
        Plot[Evaluate@{a["p"] a["f"]}, a["range"], PlotRange->Full],
        Plot[Evaluate@{Transpose[{a["p"]}].{a["p"]}}, a["range"], PlotRange->Full]
      }},
      Spacings-> 0, 
      ImageSize -> Full,
      AspectRatio -> 1/2
    ]
]


mDot[a11_] := Module[{n = DiagonalMatrix[a11["n"]], i, a = a11["alpha"], A = a11["A"], b = a11["b"], vf = a11["wVars"][[-3]]},
   i = IdentityMatrix[Length[a]];
   vf + a .(A - A. n. A). a - 2 b. (i - n .A) . a
   ];



mDot2[a11_, d_] := Module[{n = DiagonalMatrix[a11["n"]], i, a = a11["alpha"], A = a11["A"], b = a11["b"], vf = a11["wVars"][[-3]]},
   i = IdentityMatrix[Length[a]];
   d  .(A - A. n. A). d
   ];



mDot3[a11_] := Module[{n = DiagonalMatrix[a11["n"]], i, a = a11["alpha"], A = a11["A"], b = a11["b"], vf = a11["wVars"][[-3]]},
   i = IdentityMatrix[Length[a]];
   {2 a .(A - A. n. A) - 2 b. (i - n .A) ,
    Eigenvalues[A - A. n. A],
    Eigenvectors[A - A. n. A],
    A,
    a,
    b
    }
   ];



varBH[f_, a_] :=
  NIntegrate[f^2/Total[a["p"] a["n"]], a["range"]] - 
   Total[NIntegrate[f a["p"] a["n"]/ Total[a["p"] a["n"]], a["range"]]^2/a["n"]];



rejectionSample[gen_, range_, n_] := Module[{res = {}},
   While[
    Length[res] < n,
    res = Join[res, Select[gen[n], range[[1]] <= # <= range[[2]] &]]
    ];
   res[[1 ;; n]]
   ];



abSystemEstimate[a_, partSize_] := Module[
{ 
  vA, vb, b, A, alpha, gen, range, bh, Nsum, 
  rx1, rx2, 
  Aest, best, fest, pest,
  Adata, bdata, bdata2, fdata, pdata, AdataFlat, bdataFlat, bdata2Flat, fdataFlat, pdataFlat,
  asumOwenVar, prgOwenVar, asumOwenBias,
  bsumVar, asumVar, aVar, asumBias, prgVar,
  n = DiagonalMatrix[a["n"]], i = IdentityMatrix[Length[a["n"]]]
},
   alpha = a["alpha"];
   b = a["b"];
   A = a["A"];
   range = a["range"][[2 ;; 3]];
   gen := a["pRndGen"];
   Nsum = Total[a["n"]];
   
   bh = a["p"] / Total[a["p"]];
   Aest = Function[{x}, Evaluate@ ((Transpose[{bh}]. {bh}) /. Global`x -> x)];
   best = Function[{x}, Evaluate@ (a["p"] a["f"]/ Total[a["p"]]^2 /. Global`x -> x)];
   fest = Function[{x}, Evaluate@ (a["f"]/ Total[a["p"]] /. Global`x -> x)];
   pest = Function[{x}, Evaluate@ (a["p"]/ Total[a["p"]] /. Global`x -> x)];
   
   
   rx1 = Table[rejectionSample[gen[[i]], range, partSize 1000], {i, Length[b]}];
   rx2 = Table[rejectionSample[gen[[i]], range, partSize 1000], {i, Length[b]}];
   
   
   
   AdataFlat = Total[ParallelMap[
      Aest,
      rx1, {2 }
      ]];
         
   bdataFlat = Total[ParallelMap[
      best,
      rx1, {2 }
      ]];
   
   bdata2Flat = Total[ParallelMap[
      best,
      rx2, {2 }
      ]];
      
   fdataFlat = Total[ParallelMap[
      fest,
      rx1, {2 }
      ]];
      
   pdataFlat = Total[ParallelMap[
      pest,
      rx1, {2 }
      ]];
   
   Adata = Table[If[i==1, {}, Mean[ Transpose[Partition[AdataFlat, i]]]], {i, 100}];
   bdata = Table[If[i==1, {}, Mean[ Transpose[Partition[bdataFlat, i]]]], {i, 100}];
   bdata2 = Table[If[i==1, {}, Mean[ Transpose[Partition[bdata2Flat, i]]]], {i, 100}];
   fdata = Table[If[i==1, {}, Mean[ Transpose[Partition[fdataFlat, i]]]], {i, 100}];
   pdata = Table[If[i==1, {}, Mean[ Transpose[Partition[pdataFlat, i]]]], {i, 100}];
   
   Print["data ready"];
   (* a's variance*)
   aVar[partSizeArg_, bdataL_] :=
    Variance[ 
      MapThread[#1.#2 &,
       {Map[Inverse, Adata[[partSizeArg]]],   bdataL[[partSizeArg]] }
       ]
      ];
      
   (* sum of b's weighted by n variance *)
   bsumVar[partSizeArg_, bdataL_] :=
    Variance[ 
      MapThread[Total[n.#2] &,
       {Map[Inverse, Adata[[partSizeArg]]],   bdataL[[partSizeArg]] }
       ]
      ];
   
   (* sum of a's variance *)
   asumVar[partSizeArg_, bdataL_] :=
    Variance[ 
      MapThread[Total[#1.#2] &,
       {Map[Inverse, Adata[[partSizeArg]]],   bdataL[[partSizeArg]] }
       ]
      ];
            
   (* sum of a's computed from Owen's formula variance *)
   asumOwenVar[partSizeArg_, bdataL_] :=
    Variance[ 
      MapThread[Total[ LeastSquares[#1, #2]] &,
       { Nsum * Adata[[partSizeArg]] - 1 Map[Transpose[{#1}].{#1}&, pdata[[partSizeArg]]] , 
         Nsum * bdataL[[partSizeArg]] - 1 fdata[[partSizeArg]] pdata[[partSizeArg]]
        }
      ]
     ];
      
   (* optimal estimator's with estimated alphas variance *)
   prgVar[partSizeArg_, bdataL_] :=
    Mean[ 
      MapThread[
       Function[{iAe, be}, a["wVars"][[-3]] + (iAe.be).(A - A. n. A).(iAe.be) - 2 b. (i - n.A).(iAe.be)],
       {Map[Inverse, Adata[[partSizeArg]]],   bdataL[[partSizeArg]] }
       ]
      ]/partSizeArg;
      
   (* sum of a's bias *)
   asumBias[partSizeArg_, bdataL_] :=
    Mean[ 
      MapThread[
       Total[#1.#2] - Total[alpha] &,
       {Map[Inverse, Adata[[partSizeArg]]],   bdataL[[partSizeArg]] }
       ]
      ];
   
   (* sum of a's bias *)
   asumOwenBias[partSizeArg_, bdataL_] :=
    Mean[ 
      MapThread[Total[ LeastSquares[#1, #2]]- Total[alpha] &,
       { Nsum * Adata[[partSizeArg]] - 1. Map[Transpose[{#1}].{#1}&, pdata[[partSizeArg]]] , 
         Nsum * bdataL[[partSizeArg]] - 1. fdata[[partSizeArg]] pdata[[partSizeArg]]
        }
      ]
     ];
   
   <|
    "rx1" -> rx1,
    "rx2" -> rx2,
    "Adata" -> AdataFlat,
    "bdata" -> bdataFlat,
    "bdata2" -> bdata2Flat,
    "fdata"-> fdataFlat,
    "pdata" -> pdataFlat,
    "avCor" -> Function[aVar[#, bdata]],
    "asumvCor" -> Function[asumVar[#, bdata]],
    "asumbCor" -> Function[asumBias[#, bdata]],
    "prgVarCor" -> Function[prgVar[#, bdata]],
    "av" -> Function[aVar[#, bdata2]],
    "asumv" -> Function[asumVar[#, bdata2]],
    "asumb" -> Function[asumBias[#, bdata2]],
    "prgVar" -> Function[prgVar[#, bdata2]],
    "bVar" -> Function[bsumVar[#, bdata2]],    
    "asumvOwenCor" -> Function[asumOwenVar[#, bdata]],
    "asumbOwenCor" -> Function[asumOwenBias[#, bdata]],
    "resultGrid" -> 
     Grid[{
       {"A", A},
       {"A mean", Mean[AdataFlat]},
       {"b", b},
       {"b mean", Mean[bdataFlat]},
       
       {"A inverse", Inverse[A]},
       {"A mean inverse", Inverse[Mean[AdataFlat]]},
       
       {"alpha", Inverse[A].b},
       {"alpha mean", Inverse[Mean[AdataFlat]].Mean[bdataFlat]},
       
       {"alpha total", Total[Inverse[A].b]},
       {"alpha total mean", Total[Inverse[Mean[AdataFlat]].Mean[bdataFlat]]},
       
       {"A(" <> ToString[partSize] <> ") mean inverse variance",  Variance[Map[Inverse, Adata[[partSize]]]]},
       {"b(" <> ToString[partSize] <> ") mean variance",  Variance[ bdata[[partSize]]]},
       
       {"alpha(" <> ToString[partSize] <> ") variance (Correlated)", aVar[partSize, bdata]},
       {"alpha(" <> ToString[partSize] <> ") sum variance (Correlated)", asumVar[partSize, bdata]},
       
       {"alpha(" <> ToString[partSize] <> ") variance", aVar[partSize, bdata2]},
       {"alpha(" <> ToString[partSize] <> ") sum variance", asumVar[partSize, bdata2]},
       
       {"result(" <> ToString[partSize] <> ") mean variance", Variance[fdata[[partSize]]]},
       {"result(" <> ToString[partSize] <> ") mean", Mean[fdata[[partSize]]]},
       {"result All", Mean[fdataFlat]}
       },
      Frame -> All
      ]
    |>
   ];


pltAANA[a_]:=Module[{
 cs, v, vv
},
 cs = Transpose @(mDot3 @ a)[[3]];

 GraphicsGrid @
 {{
 ParametricPlot3D[
  Normalize[{Sin[i]Sin[j],Cos[i]Sin[j], Cos[j]}] 
    mDot2 @@ Append[{a}, cs.Normalize[{Sin[i]Sin[j],Cos[i]Sin[j], Cos[j]}]],
  {i, 0, 2Pi},{j, Pi, -Pi}, PlotRange->Full, AspectRatio->1/1.2
 ],
 Show[
 ParametricPlot3D[
  Normalize[{Sin[i]Sin[j],Cos[i]Sin[j], Cos[j]}] 
    mDot2 @@ Append[{a}, Normalize[{Sin[i]Sin[j],Cos[i]Sin[j], Cos[j]}]],
  {i, 0, 2Pi},{j, Pi, -Pi}, PlotRange->Full, AspectRatio->1/1.2
 ],
 ListPointPlot3D[
  {Normalize[a["alpha"]]
    mDot2[a, a["alpha"]]}
 ]
 ]
 }}
];



pltVar[aa_, k_, ul_ : 100, mtex_: (#1&)] := 
Block[{aaa, colors, styles, aspRatio},
  aaa = abSystemEstimate[aa, k];
  colors = Lighter/@{Red, Green, Blue, Black, Purple};
  colors = RGBColor[#1]&/@ StringSplit["#E69F00 #56B4E9 #009E73 #F0E442 #0072B2"];
  styles = Flatten[ {{Thick, #1}, {Thick, Dashed, #1}}& /@ colors, 1];
  aspRatio = 1/1;
  (*Print[aaa["resultGrid"]];*)
  Grid[{
  {
   LineLegend[
     colors,
     mtex/@{"Direct", "Direct, uncorrelated", "Proggresive", "Proggresive, uncorrelated"}, 
     LegendLayout->{"Column", 4}, 
     LabelStyle-> Small,
     LegendMargins->0
     ]
  },
  {GraphicsGrid[{
     (* modify aspect ratio of all plots in the list below ..*)
     Show[#1, AspectRatio -> aspRatio, BaselinePosition -> Axis]& /@
     {
     (* balance and power heuristics overlays*)
     (*Show[
      ListPlot[{
        Table[aaa["asumvCor"][i], {i, 2, ul}],
        Table[aaa["asumv"][i], {i, 2, ul}],
        Table[aaa["prgVar"][i], {i, 2, ul}],
        Table[aaa["prgVarCor"][i], {i, 2, ul}]
        }, 
       PlotRange -> Automatic, 
       PlotStyle -> Evaluate @ ({Thick, #1}& /@ styles[[{1, 2, 5}]]),
       AxesOrigin-> {0, Evaluate @ aa["wVars"][[-1]]},
       Joined -> True
       ],
      Plot[Evaluate@aa["wVars"][[-3 ;; -2]], {x, 0, ul},
       PlotStyle -> Dashed,
       PlotLabels -> Placed[{"balance var", "power var"}, {Left}]
       ]
      ],*)
     (* variance*)
     Show[
      LogLogPlot[Evaluate@(aa["wVars"][[{-3, -1}]]/x/1.2), {x, 0, ul},
       PlotRange->{{2, Automatic}, Automatic},
       PlotStyle -> {{Thickness[Medium], Dashed, Black}},
       PlotLabels -> Placed[mtex/@{"balance var", "optimal var"}, {Right}]
       ],
      ListLogLogPlot[{
        Table[aaa["asumvCor"][i], {i, 2, ul}],
        Table[aaa["asumv"][i], {i, 2, ul}],
        Table[aaa["prgVarCor"][i], {i, 2, ul}],
        Table[aaa["prgVar"][i], {i, 2, ul}]
        }, 
       PlotRange -> Automatic, 
       PlotStyle -> Evaluate @ ({Thick, #1}& /@ styles[[{1, 2, 3, 4}]]),
       (*AxesOrigin-> {0, Evaluate @ aa["wVars"][[-1]]},*)
       Joined -> True
       ]
      ],
     (*  bias *)
     ListPlot[{
       Table[aaa["asumbCor"][i], {i, 2, ul}],
       Table[aaa["asumb"][i], {i, 2, ul}]
       }, 
      PlotRange -> Automatic,
      Joined -> True,
      PlotStyle -> Evaluate @ ({Thin, #1}& /@ styles[[{1, 2, 5}]])
      ]
     }},
     Spacings-> 0, 
     ImageSize -> Full,
     ItemAspectRatio -> aspRatio
   ]}
   }]
  ]


End[];

EndPackage[];
